package be.iminds.ilabt.jfed.restapi.resource;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserFactory;
import be.iminds.ilabt.jfed.restapi.JFedRestApiConfiguration;
import be.iminds.ilabt.jfed.restapi.json.Experiment;
import be.iminds.ilabt.jfed.restapi.json.SpeaksForCredential;
import be.iminds.ilabt.jfed.restapi.json.User;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.KeyUtil;
import be.iminds.ilabt.util.jsonld.UriTool;
import com.codahale.metrics.annotation.Timed;
import com.fasterxml.jackson.annotation.*;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.*;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.List;

/**
 * RootResource exposes basic config info, and links to the other resources
 */


@Path("/root")
@Produces(MediaType.APPLICATION_JSON)
public class RootResource {
    private static final Logger LOG = LoggerFactory.getLogger(RootResource.class);

    @Context
    UriInfo uriInfo;

    private final JFedRestApiConfiguration configuration;

    private final String speakingForX509Certificate;
    private final String speakingForUserName;
    private final String speakingForUserUrn;

    public RootResource(JFedRestApiConfiguration configuration) {
        this.configuration = configuration;

        //noinspection ConstantConditions
        if (configuration.getSfaUserLoginPemFilename() == null) {
            LOG.error("Server config error: sfaUserLoginPemFilename option missing");
            throw new WebApplicationException("Server config error: sfaUserLoginPemFilename option missing", Response.Status.INTERNAL_SERVER_ERROR);
        }

        try {
            String pemFileContent = IOUtils.fileToString(configuration.getSfaUserLoginPemFilename());
            GeniUserFactory geniUserFactory = new GeniUserFactory(null);
            GeniUser geniUser = geniUserFactory.createGeniUser(pemFileContent, null, null, null);
            this.speakingForX509Certificate = KeyUtil.x509certificateChainToPem(geniUser.getClientCertificateChain());
            this.speakingForUserName = geniUser.getUserUrn().getResourceName();
            this.speakingForUserUrn = geniUser.getUserUrnString();
        } catch (Exception e) {
            throw new RuntimeException("Failed to extract info on speaking user", e);
        }
    }

    @JsonInclude(JsonInclude.Include.NON_NULL)
    @JsonIgnoreProperties(value={"@context"})
    public static class Root {
        private final URI experiment;
        private final URI speaksFor;
        private final URI user;
        private final URI debugData;

        private final URI self;

        private final String callerIp;
        private final String callerUserUrn;

        private final String speakingForX509Certificate;
        private final String speakingForUserName;
        private final String speakingForUserUrn;

        public Root(@Nonnull UriInfo uriInfo,
                    @Nonnull JFedRestApiConfiguration configuration,
                    @Nullable String callerIp,
                    @Nullable String callerUserUrn,
                    @Nullable String speakingForX509Certificate,
                    @Nullable String speakingForUserName,
                    @Nullable String speakingForUserUrn) {
            UriTool uriTool = configuration.getUriTool();
            this.experiment = uriTool.getUriUsingFedmonObject(Experiment.class, null);
            this.speaksFor = uriTool.getUriUsingFedmonObject(SpeaksForCredential.class, null);
            this.user = uriTool.getUriUsingFedmonObject(User.class, null);
            this.debugData = uriInfo.getBaseUriBuilder().path(DebugDataResource.class).build();

            this.callerIp = callerIp;
            this.callerUserUrn = callerUserUrn;

            this.speakingForX509Certificate = speakingForX509Certificate;
            this.speakingForUserName = speakingForUserName;
            this.speakingForUserUrn = speakingForUserUrn;

            String baseUrl = uriTool.getBaseUrlString();
            assert baseUrl != null;
            assert baseUrl.endsWith("/");
            try {
                this.self = new URI(baseUrl+"");
            } catch (URISyntaxException e) {
                throw new RuntimeException("Failed to initialise using baseUrl=\""+baseUrl+"\"", e);
            }
        }

        @JsonCreator
        public Root(@JsonProperty("experiment") URI experiment,
                    @JsonProperty("speaksFor") URI speaksFor,
                    @JsonProperty("user") URI user,
                    @JsonProperty("debugData") URI debugData,
                    @JsonProperty("self") URI self,
                    @JsonProperty("callerIp") String callerIp,
                    @JsonProperty("callerUserUrn") String callerUserUrn,
                    @JsonProperty("callerUserUrn") String speakingForX509Certificate,
                    @JsonProperty("speakingForUserName") String speakingForUserName,
                    @JsonProperty("speakingForUserUrn") String speakingForUserUrn) {
            this.experiment = experiment;
            this.speaksFor = speaksFor;
            this.user = user;
            this.debugData = debugData;
            this.self = self;
            this.callerIp = callerIp;
            this.callerUserUrn = callerUserUrn;
            this.speakingForX509Certificate = speakingForX509Certificate;
            this.speakingForUserName = speakingForUserName;
            this.speakingForUserUrn = speakingForUserUrn;
        }

        @JsonProperty
        public URI getExperiment() {
            return experiment;
        }

        @JsonProperty
        public URI getSpeaksFor() {
            return speaksFor;
        }

        @JsonProperty
        public URI getUser() {
            return user;
        }

        @JsonProperty
        public URI getDebugData() {
            return debugData;
        }

        @JsonProperty
        public String getSpeakingForX509Certificate() {
            return speakingForX509Certificate;
        }

        @JsonProperty
        public String getSpeakingForUserName() {
            return speakingForUserName;
        }

        @JsonProperty
        public String getSpeakingForUserUrn() {
            return speakingForUserUrn;
        }

        @JsonIgnore
        public String getCallerIp() {
            return callerIp;
        }

        @JsonIgnore
        public String getCallerUserUrn() {
            return callerUserUrn;
        }

        @Nonnull
        @JsonProperty("caller")
        public Caller getCaller() {
            return new Caller(callerIp, callerUserUrn);
        }

        @JsonProperty("@id")
        public URI getSelf() {
            return self;
        }

        @Override
        public String toString() {
            ObjectMapper mapper = new ObjectMapper();
            try {
                return mapper.writerWithDefaultPrettyPrinter().writeValueAsString(this);
            } catch (JsonProcessingException e) {
                LOG.error("Exception converting Admin Info to JSON", e);
                return "Exception converting Admin Info to JSON: "+e.getMessage();
            }
        }
    }

    @JsonInclude(JsonInclude.Include.NON_NULL)
    @JsonIgnoreProperties(value={"@context"})
    private static class Caller {
        private final String ip;
        private final String urn;

        @JsonCreator
        public Caller(
                @JsonProperty("ip") String ip,
                @JsonProperty("urn") String urn) {
            this.ip = ip;
            this.urn = urn;
        }

        @JsonProperty("ip")
        public String getIp() {
            return ip;
        }

        @JsonProperty("urn")
        public String getUrn() {
            return urn;
        }
    }

    @GET
    @Timed
    public Root get(@Context HttpServletRequest request) {
        GeniUrn authenticatedUser = request == null ? null : configuration.getAuthenticatedUser(request);
        return new Root(
                uriInfo,
                configuration,
                request == null ? null : request.getRemoteAddr(),
                authenticatedUser == null ? null : authenticatedUser.toString(),
                speakingForX509Certificate,
                speakingForUserName,
                speakingForUserUrn
        );
    }
}
